/*            */
/* Steven Bickerton */
/* Dept. of Physics/Astronomy, McMaster University */
/* bick@physics.mcmaster.ca*/
/* Made with makeScript, Thu Oct 19, 2006  18:24:42 DST */
/* Host: kuiper */
/* Working Directory: /1/home/bickersj/sandbox/Canalysis  */


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "util.h"

// --------------------------------------------------------------------
void errQuit (char *msg, ...) {

  char s[MAX_LINE_LENGTH];

  va_list args;
  va_start(args,msg);
  vsprintf(s, msg, args);
  va_end(args);

  if (errno)
    perror(s);
  else 
    fprintf(stderr, s);

  exit(EXIT_FAILURE);
}

/* -------------------------------------------------------------- */
int cmp(const void *v1, const void *v2) {  
  int i;
  double diff = *(double *)v1 - *(double *)v2;

  i = (diff > 0 ) ? 1 : -1;
  if (diff == 0) i = 0;
  return  i;  
}

// --------------------------------------------------------------------
FILE *openfile (char *path, char *mode) {

  FILE *fp;

  if ( path == NULL || path[0] == '-'  ) {
    return stdin;
  } else if ( (fp = fopen(path, mode)) != NULL)
    return fp;
  else
    errQuit("open file (%s)", path);

  return NULL;  // never reached ... surpresses compiler error
}

// ----------------------------------------------------------------------
void closefile (FILE *stream) {
  if (fclose(stream) != 0)
    errQuit("close file");
}



/* ---------------------------------------------------------------- */
/* take a string and split it into pieces at 'tok' delimiters      */
/* put the pieces into an array and return the number found       */
int split (char *s, char *tok, char *entries[], int n) {
  
  int i;

  entries[0] = strtok(s, tok);
  for(i=1;i<n;i++)
    if ( (entries[i]=strtok(NULL,tok)) == NULL ) {
      i++;
      break;
    }

  return i;
}


int intSplit (char *s, char *tok, int entries[], int n) {

  int i,n2;
  char *char_entries[n];
  n2 = split(s, tok, char_entries, n);
  for (i=0;i<n2;i++)
    entries[i] = atoi(char_entries[i]);
  
  return n2;
}

double doubleSplit (char *s, char *tok, double entries[], int n) {

  int i,n2;
  char *char_entries[n];
  n2 = split(s, tok, char_entries, n);
  for (i=0;i<n2;i++)
    entries[i] = atof(char_entries[i]);
  
  return n2;
}

// --------------------------------------------------------------------
// read in all lines and columns of data into an array -- allocate mem too
void loadFile(char *path, ARRAY *datap)
{
  int i, nlines, ncolumns, ncolumns_max, nchars, nalnum;
  char paramLine[MAX_LINE_LENGTH]; // force 1 leading white space
  char *paramtmp;
  char *space = " ";
  double value;
  double *tmp;

  // initialize some array values.
  for (i=0; i<MAX_COLUMNS; i++) {
    datap->n[i] = 0;
    datap->ndef[i] = 0;
    datap->x[i] = NULL;
    datap->sum[i] = 0.0;
    datap->sumsqr[i] = 0.0;
  }

  // read in the whole line
  FILE *fp = openfile(path, "r");
  nlines = ncolumns_max = 0;
  while ( fgets(paramLine, MAX_LINE_LENGTH, fp) != NULL ) {
    
    // get rid of any comment lines
    if (paramLine[0]=='#')
      continue;
    
    // lose trailing whitespace, replace tabs with spaces
    nchars = (int) strlen(paramLine);
    nalnum = 0;
    for (i=nchars-1; i>=0; i--) {

      // set trailing white-space to \0
      if ( isspace(paramLine[i]) && ! nalnum) {
	paramLine[i] = '\0';
      } else {
	nalnum++;
      }
      
      // replace each tab character with a space
      if ( paramLine[i] == '\t' )
	paramLine[i] = space[0];
    }
    
    // if the whole line was space skip it
    if ( ! nalnum ) 
      continue;
    
  
    // split on spaces
    ncolumns = 0;
    while ( (paramtmp=strtok( (ncolumns?NULL:paramLine), space)) != NULL ) {
      
      // get more memory for this column and initialize to NAN
      if ( ( tmp=realloc(datap->x[ncolumns], 
			 (size_t) (nlines+1)*sizeof(double)) ) == NULL )
	errQuit("memory allocation");
      datap->x[ncolumns] = tmp;
      datap->x[ncolumns][nlines] = NAN;  // bad data flag
      datap->n[ncolumns] = nlines+1;     // new array size

      // record the value, the number of valid entries and some info
      value = atof(paramtmp);
      if ( !isnan(value) && !isinf(value) ) {
	datap->x[ncolumns][nlines] = value;
	datap->ndef[ncolumns]++;
	if ( value < datap->min[ncolumns] || ! nlines )
	  datap->min[ncolumns] = value;
	if (value > datap->max[ncolumns]  || ! nlines )
	  datap->max[ncolumns] = value;
	datap->sum[ncolumns] += value;
	datap->sumsqr[ncolumns] += value*value;
      }
 
      ncolumns++;
    }

    if (ncolumns > ncolumns_max)
      ncolumns_max = ncolumns;
    

    nlines++;
  }
  closefile(fp); 
  
  // if ncolumns_max is still zero, we got nothin ...
  if ( ! ncolumns_max )
    fprintf(stderr, "Warning:  No data found in file: %s\n", path);
  
  datap->columns = ncolumns_max;
}



long loadCoords (COORD **coord, char *filename, int xcol, int ycol) {

  ARRAY data;
  ARRAY *datap = &data;
  long i,n;

  xcol--;
  ycol--;

  loadFile(filename, datap);

  // the number of entries is shortest column
  n = datap->ndef[xcol];
  if (datap->ndef[ycol] < n)
    n = datap->ndef[ycol];

  if ( (*coord = malloc(n*sizeof(COORD)) ) == NULL )
    errQuit("memory allocation in loadCoords()\n");

  for(i=0;i<n;i++) {
    (*coord)[i].x = datap->x[xcol][i];
    (*coord)[i].y = datap->x[ycol][i];
  }
  
  freeArray2D(datap);

  return n;
}

// --------------------------------------------------------------

void printArray2D (ARRAY *datap) {

  long i, j;

  for (i=0; i<datap->n[0]; i++) {
    for (j=0; j<datap->columns; j++) 
      printf("%+12.12g ", datap->x[j][i]);
    printf ("\n");
  }
}

void freeArray2D (ARRAY *datap) {

  long i;
  for (i=0; i<datap->columns; i++)
    free(datap->x[i]);

}

