#!/usr/bin/env perl
#
# original filename: phasePlot.pl
#
# Steven Bickerton
#  Dept. of Physics/Astronomy, McMaster University
#  bick@physics.mcmaster.ca
#  Made with makeScript, Thu Aug 16, 2007  13:28:50 DST
#  Host: kuiper
#  Working Directory: /1/home/bickersj/working/TASS
#


use strict;
use warnings;
use File::Basename;

my $exe = basename($0);
my $usage = "Usage: $exe tsfile [force] [dump] [{eps|png}file]\n";

my ($tsfile, @args) = @ARGV;
die $usage unless $tsfile;

my ($force,$dumping, $pngfile, $epsfile) = (0,0, "", "");
foreach my $arg (@args) {
    $force += 1 if $arg=~/^f.*/;
    $dumping += 1 if $arg=~/^d.*/;
    $pngfile = $arg if $arg=~/.*\.png$/;
    $epsfile = $arg if $arg=~/.*\.eps$/;
}

my ($dir, $ID, $ext) = $tsfile =~ /(\/?.*\/)?(\d+)(\..*)\.ts/;
$dir = "" unless $dir;
$ext = "" unless $ext;
my $pdmfile = "${dir}${ID}${ext}.pdm";
my $vphasefile = "${dir}${ID}${ext}.ph1";
my $iphasefile = "${dir}${ID}${ext}.ph2";
my $vsmthfile = "${dir}${ID}${ext}.smth1";
my $ismthfile = "${dir}${ID}${ext}.smth2";

# tsfile columns
my ($tcol, $vcol, $icol)  = (5, 7, 11);
# pdmfile columns
my ($pcol, $pdmcol) = (2, 7);
my ($pdmthresh) = (100);
# phase columns
my ($ph_tcol, $ph_ycol, $ph_ocol) = (1,2,4);


# get the peaks and best period from the pdm file
my @lines = split /\n/, `getPeaks $pdmfile $pdmthresh c${pcol}:${pdmcol}`;
die "No peaks above $pdmthresh\n" unless @lines;
my @peaks = ();
foreach my $line ( @lines ) {
    push @peaks, [ split /\s+/,$line ];
}

@peaks = sort {$a->[1] <=> $b->[1]} @peaks;
my $period = $peaks[-1]->[0];



# phase the data.
(-s $vphasefile && ! $force ) || system("phaseFold $tsfile $period ${tcol}:${vcol} > $vphasefile");
(-s $iphasefile && ! $force ) || system("phaseFold $tsfile $period ${tcol}:${icol} > $iphasefile");


# smooth the data.
my ($gauss, $width) = (1, 0.1*$period);
(-s $vsmthfile && ! $force ) || system("smoothC $vphasefile ${ph_tcol}:${ph_ycol} $gauss $width > $vsmthfile");
(-s $ismthfile && ! $force ) || system("smoothC $iphasefile ${ph_tcol}:${ph_ycol} $gauss $width > $ismthfile");

# make plot script
my $setterm = "set term ";
if ($pngfile) {
    $setterm .= "png enhanced";
} elsif ($epsfile) {
    $setterm .= "postscript eps enhanced color";
} else {
    $setterm .= "x11 enhanced\n";
}

my $setoutput = "";
$setoutput = "set out \'$pngfile\'" if $pngfile;
$setoutput = "set out \'$epsfile\'" if $epsfile;

my $gp_script = "#!/usr/bin/env gnuplot\n".
    "#\n".
    "$setterm\n".
    "$setoutput\n".
    "set view map\n".
    "set grid\n".
    "XSIZE=1.0\n".
    "YSIZE=1.0\n".
    "set size XSIZE,YSIZE\n".
    "MX=0.0*XSIZE\n".
    "MY=0.05*YSIZE\n".
    "PX=1.0*XSIZE\n".
    "PY=0.55*YSIZE\n".
    "set title \"$tsfile   P=$period\"\n".
    "set style line 1 pt 7 ps 2\n".
    "set style line 2 lt 2 lw 2 lc rgb \"green\"\n".
    "set style line 3 lt 2 lw 2 lc rgb \"red\"\n".
    "set xrange [0:2.0*$period]\n".
    "set x2range [0:2]\n".
    "set xtic nomirror\n".
    "set x2tic 0.2 nomirror\n".
    "set yrange [:] reverse\n".
    "set arrow 1 from $period, gr 0 to $period, gr 1 nohead\n".
    "set multiplot\n".
    "set bmargin 0\n".
    "set tmargin 0\n".
    "set lmargin 0\n".
    "set rmargin 0\n".
    "set format x \"\"\n".
    "set size PX, PY\n".
    "set origin MX,PY-1.5*MY\n".
    "set ylabel \"V (mag)\" offset -1,0\n".
    "splot \'$vsmthfile\' using 1:2:(0) t \"\" w l ls 2,\\\n".
    "\'$vphasefile\' using ${ph_tcol}:${ph_ycol}:${ph_ocol} t \"\" w p pt 7 ps 2 palette\n".
    "\n".
    "unset title\n".
    "set origin MX,MY\n".
    "set format x \"%%.f\"\n".
    "set ylabel \"I (mag)\" offset -1,0\n".
    "set xlabel \"Time (days)\"\n".
    "splot \'$ismthfile\' using 1:2:(0) t \"\" w l ls 3,\\\n".
    "\'$iphasefile\' using ${ph_tcol}:${ph_ycol}:${ph_ocol} t \"\" w p pt 7 ps 2 palette\n";

# plot

open(GNUPLOT, "|gnuplot -persist");
printf GNUPLOT $gp_script;
printf STDOUT "$gp_script" if $dumping;
close(GNUPLOT);

exit 0;
