#!/usr/bin/env perl
#
# original filename: extractStars.pl
#
# Steven Bickerton
#  Dept. of Physics/Astronomy, McMaster University
#  bick@physics.mcmaster.ca
#  Made with makeScript, Mon Jul  9, 2007  14:17:22 DST
#  Host: unknown
#  Working Directory: /Users/bick/working/TASS
#
# Recalibrate the photometry for TASS .cal file
# - data are binned by JD, RA, and Dec (a 3rd order tensor structure).
# - the mean mag of each star is computed
# - in a cellsize x cellsize box, for a given JD, the amount by which each 
#    star's mag deviates is computed and the median is taken.
#    --> this is the calibration error for that observation, in that sky region
# - in each cellsize x cellsize box, all stars have the correction applied.
#
# - this method leaves partial-cells hanging off the field, so right and top
#   edges have smaller cells (this could stand to be changed).
#

use strict;
use warnings;
use File::Basename;

use Time::HiRes qw(gettimeofday);

my $exe = basename($0);
my $usage = "Usage: $exe infile cellsize\n";

my ($infile, $cellsize) = @ARGV;
die $usage unless $infile and -f $infile;
die $usage unless $cellsize;

# the filenames for the v and i declination corrections
# and the columns containing declination, and the correction
my ($vcorrfile, $icorrfile) = ("corrections_v.out", "corrections_i.out");
my ($cdecCorr, $ccorrCorr) = (2,9);

# column definitions for the .cal file
#   if you use a different input file format, change these accordingly
#
# needed: all of them must be defined (i store tmp values in the 
#         otherwise-unused spaces to avoid allocating new memory.
my ($cID, $cN, $cRA, $cDec, $cjd, 
    $cf1, $cm1, $cmer1, $cq1, 
    $cf2, $cm2, $cmer2, $cq2) = (0 .. 12);


my $nsigma = 4.0;  # globals for meanClip
my $outbase = $infile;
my $outfile = "${outbase}.rcal";

# to turn on profiling (process timing) set an env. variable TASSPROFILE to 1
#   ... or hardcode $profile to a non-zero value
my $profile = ($ENV{"TASSPROFILE"}) ? (1) : 0;

# >= $Nmin stars must be in a time-series to include it
# in the computation of the median
my $Nmin = 5;  

sub medianInsitu ($);
sub meanClip ($);
			      
#######################  profiling
my ($tA,$tB,$t0);
$tA = $tB = $t0 = gettimeofday();
printf STDERR "%-20s", "Loading data:" if $profile;
#######################


#########################################################################
# load the known declination corrections   ##############################
my %Vcorr;
open (VCORR, "$vcorrfile");
while(<VCORR>) {
  next if /^\#/;
  my @line = split;
  $Vcorr{$line[$cdecCorr]} = $line[$ccorrCorr];
}
close (VCORR);

my %Icorr;
open (ICORR, "$icorrfile");
while(<ICORR>) {
  next if /^\#/;
  my @line = split;
  $Icorr{$line[$cdecCorr]} = $line[$ccorrCorr];
}
close (ICORR);

# assign values in 0.01 degrees
my $vcorr_last = $Vcorr{"-10.00"};
my $icorr_last = $Icorr{"-10.00"};
foreach my $dec (map {sprintf "%.2f", $_/100.0} (-1000 .. 10000)) {
  if ($Vcorr{$dec}) {
    $vcorr_last = $Vcorr{$dec};
  } else {
    $Vcorr{$dec} = $vcorr_last;
  }

  if ($Icorr{$dec}) {
    $icorr_last = $Icorr{$dec};
  } else {
    $Icorr{$dec} = $icorr_last;
  }
}



##########################################################################
#
# load the data                    #######################################
#
##########################################################################
my %IDs;      # the index numbers for each ID
my $nID = 0;
my @data;
my %JDs;      # the index numbers for each JD
my ($ra_min,$ra_max, $dec_min, $dec_max) = (360.0, 0.0, 90.0, -90.0);
open(INFILE, "$infile");
while (<INFILE>) {
  next if /^\#/;

  my @record = (split);
  my ($ID,$ra,$dec,$jd) = @record[$cID,$cRA,$cDec,$cjd];

  # need the dimensions to handle cells
  $ra_min  = $ra  if ( $ra  < $ra_min  );
  $ra_max  = $ra  if ( $ra  > $ra_max  );
  $dec_min = $dec if ( $dec < $dec_min );
  $dec_max = $dec if ( $dec > $dec_max );

  # need to know which JDs we have
  $JDs{$jd} = 1;

  # apply the declination correction as we read in the magnitudes
  my $decS = sprintf "%.2f", $record[$cDec];
  $record[$cm1]  += $Vcorr{$decS};
  $record[$cm2]  += $Icorr{$decS};
  
  # I won't use 'V' and 'I' labels, so I'll store numbers in there place
  #   to save some memory
  $record[$cf1] = 0.0;   # this will hold mean1 (the mean V magnitude)
  $record[$cf2] = 1.0;   #         "      mean2 ( "   "   I  "       )

  $IDs{$ID} = $nID++ unless $IDs{$ID};

  # data holds references to the arrays of records for each target
  push @{ $data[ $IDs{$ID} ] }, \@record;

}
close(INFILE);

my $nRA = int(($ra_max-$ra_min)/$cellsize);       # no. of RA cells
my $nDec = int(($dec_max-$dec_min)/$cellsize);    # no. of Dec cells


# make a hash of JD values and the array index number they correspond to.
my @JDs = sort {$a<=>$b} keys %JDs;
my $nJD = 0;
foreach my $jd (@JDs) {   $JDs{$jd} = $nJD++;  }
 


############################################################################
# 
# go through and get the weighted-mean magnitudes for each ID  
#  we'll use the weighted-mean to compute the magnitude deviations
#
############################################################################

################ profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "Wmean:" if $profile;
$tA = $tB;
##################

foreach my $IDi ( 0 .. $#data ) {
    
    #my $IDi = $IDs{$ID};  # the index for this ID
    my ($mean1, $mean2, $w1sum, $w2sum, $n1, $n2) = (0,0,0,0,0,0);
  RECORD: foreach my $ref (@{ $data[ $IDi ] }) {

      my ($ra, $dec, $jd, $mag1, $merr1, $q1, $mag2, $merr2, $q2) = 
	  @{$ref}[$cRA,$cDec,$cjd,$cm1,$cmer1,$cq1,$cm2,$cmer2,$cq2];
      
      next RECORD 
	  if ($merr1 == 0 || $merr2 == 0 || 
	      $mag1 > 25  || $mag2 > 25  || 
	      $q1 || $q2);
      
      my $w1 = 1.0/$merr1**2;
      $mean1 += $w1 * $mag1;
      $w1sum += $w1;
      $n1++;
      my $w2 = 1.0/$merr2**2;
      $mean2 += $w2 * $mag2;
      $w2sum += $w2;
      $n2++;

      # put the jd index number in N since we're not using that
      $ref->[$cN] = $JDs{$jd};
      
  }
    
    $w1sum = 1 if $n1 == 0;
    $w2sum = 1 if $n2 == 0;

    $mean1 /= $w1sum;
    $mean2 /= $w2sum;

    # put the new mean values in place of the V and I labels of the first obs
    $data[$IDi]->[0]->[$cf1] = $mean1;
    $data[$IDi]->[0]->[$cf2] = $mean2;

    
}


############################################################################
# 
####   go through and get the median mag difference for each JD  #########
#
############################################################################

#############################
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "MagDiff arrays:"  if $profile;
$tA = $tB;
#############################

my @celljdMagDiffs;
foreach my $IDi (0 .. $#data ) {

    my $N = scalar @{ $data[$IDi] };
    next unless $N > $Nmin;

    # these values were stuffed in place of V, I labels to avoid
    #  making a whole new structure and consuming memory
    my ($smean1,$smean2) = @{ $data[$IDi]->[0] }[$cf1,$cf2];
    
  RECORD: foreach my $ref (@{$data[$IDi]}) {

      my ($ID, $jdi, $ra, $dec, $jd, $mag1, $merr1, $q1, $mag2, $merr2, $q2) = 
	  @{$ref}[$cID,$cN,$cRA,$cDec,$cjd,$cm1,$cmer1,$cq1,$cm2,$cmer2,$cq2];
      
      next RECORD 
	  if ($merr1 == 0 || $merr2 == 0 || 
	      $mag1 > 25  || $mag2 > 25  || 
	      $q1 || $q2);
      

      my $xcell = sprintf "%d", ($ra-$ra_min)/$cellsize;
      my $ycell = sprintf "%d", ($dec-$dec_min)/$cellsize;
      
      my ($mag1diff,$mag2diff) = ($mag1-$smean1, $mag2-$smean2);

      # uncomment this line to dump the magnitude deviations to stdout
      # printf STDOUT "$jd $mag1diff $mag2diff\n";

      foreach my $i (0) { # (-1, 0, 1) {
	  foreach my $j (0) { #(-1, 0, 1) {

	      my $xi = $xcell + $i;
	      my $yi = $ycell + $j;
	      
	      # check bounds and exclude corner cells
	      next if ($xi<0 || $yi<0 || $xi>$nRA || $yi>$nDec);

	      push @{$celljdMagDiffs[$xi][$yi][$jdi][0]}, $mag1diff;
	      push @{$celljdMagDiffs[$xi][$yi][$jdi][1]}, $mag2diff;

	  }
      }
      
  }
}



###################################################################
# compute the medians for each JD/region data set
#  put the answer in place of the of the array reference to the dataset.
#
#################################################################
############################ profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "MagDiff med:"  if $profile;
$tA = $tB;
############################

# for each cell (xi,yi), for each observation (jdi), for each filter (i)
#  get the median
foreach my $xi (0 .. $nRA) {
    foreach my $yi (0 .. $nDec) {
 	foreach my $jdi (0 .. $nJD) {
	    foreach my $i (0,1) {

		next unless $celljdMagDiffs[$xi][$yi][$jdi][$i];

		$celljdMagDiffs[$xi][$yi][$jdi][$i] = 
		    medianInsitu( $celljdMagDiffs[$xi][$yi][$jdi][$i] );
		#$celljdMagDiffs[$xi][$yi][$jdi][$i] = 
		#    meanClip( $celljdMagDiffs[$xi][$yi][$jdi][$i] );

	    }
 	}
    }
}


 
############################################################################
# 
######  go through each ID and adjust each mag by the median deviation
#          for it's JD/region
#
############################################################################

#############################  profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "Recalibration:" if $profile;
$tA = $tB;
############################

foreach my $IDi (0 .. $#data) {

  RECORD1: foreach my $ref (@{$data[$IDi]}) {

      my ($jdi, $ra, $dec, $jd, $mag1, $merr1, $q1, $mag2, $merr2, $q2) = 
	  @{$ref}[$cN,$cRA,$cDec,$cjd,$cm1,$cmer1,$cq1,$cm2,$cmer2,$cq2];
      
      next RECORD1 
	  if ($merr1 == 0 || $merr2 == 0 || 
	      $mag1 > 25  || $mag2 > 25  || 
	      $q1 || $q2);

      my ($xcell, $ycell) = 
	  map {sprintf "%d", $_/$cellsize} ($ra-$ra_min, $dec-$dec_min);

      my $mag1corr = $celljdMagDiffs[$xcell][$ycell][$jdi][0];
      my $mag2corr = $celljdMagDiffs[$xcell][$ycell][$jdi][1];
      
      # set the bad quality flags if no mag correction was defined
      if (! $mag1corr) {
	  $ref->[8]  = 1;
	  $mag1corr = 0.0;
      }
      if (! $mag2corr) {
	  $ref->[12] = 1;
	  $mag2corr = 0.0;
      }

      # sub in the corrected magnitudes
      $ref->[$cm1] = $mag1-$mag1corr;
      $ref->[$cm2] = $mag2-$mag2corr;

  }
}



##########################################################################
#
### output all of it to files  #########################################
#
#########################################################################

################################## profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "output:" if $profile;
$tA = $tB;
#################################

open(OUTFILE, ">${outfile}");
my @col;
@col[$cID,$cN,$cRA,$cDec,$cjd,
     $cf1, $cm1, $cmer1, $cq1, 
     $cf2, $cm2, $cmer2, $cq2] = 
    ("ID","N","RA","Dec","jd",
     "f1"," m1"," mer1"," q1", 
     "f2"," m2"," mer2"," q2");

foreach my $i (0 .. 12) { printf OUTFILE "# %02d %s\n", $i, $col[$i]; }
foreach my $IDi (0 .. $#data ) {
    foreach my $ref ( @{$data[$IDi]} ) {
	printf OUTFILE "%d %3d  %8.4f %8.4f %13.5f ".
	    "  %1s %6.3f %6.3f %1d  ".
	    "%1s %6.3f %6.3f %1d\n", @$ref;
    }
}
close(OUTFILE);


################################  profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
$tA = $tB;
################################

exit 0;


############################################################################
# meanClip() can be used in place of medianInsitu() - it's faster, but not
#    as robust
# eg.
# my $meanClip = meanClip(\@array);
#  ... where @array is (x0, x1, x2, x3, ...)
#
sub meanClip($) {

    my ($data_ref) = @_;
    my ($mean,$sumSqr) = (0.0,0.0);
    my $n = scalar @$data_ref;

    # first pass
    foreach my $x (@$data_ref) {
	$mean += $x;
	$sumSqr += $x*$x;
    }
    $mean /= $n;
    my $rms = sqrt($sumSqr/$n - $mean*$mean);
    my $limit = $nsigma*$rms;

    # second pass with clipping
    ($mean,$n) = (0.0,0);
    foreach my $x (@$data_ref) {
	if ( $x> -$limit  &&  $x<$limit ) {
	    $mean += $x;
	    $n++;
	}
    }

    return ($n) ? ($mean/$n) : 0.0;
}

########################################################################
# medianInsitu() is called to get the median magnitude deviation for a
#    time/region.  
# usage:
# my $median = medianInsitu(\@array);
#   ... with @array = (x0, x1, x2, ...)
#
# it is 'insitu' because the 'sort' is done insitu, rather than by copying
#   this uses less memory
sub medianInsitu($) {
    my ($data_ref) = @_;

    @$data_ref = sort {$a<=>$b} @$data_ref;

    my $n = scalar @$data_ref;
    my $nBy2 = int ($n / 2);

    return 0 if ($n==0);

    # if odd take the middle value, if even avg the two middle values
    my $median;
    if ( $n % 2 == 0 ) {
	$median = ($$data_ref[$nBy2] + $$data_ref[$nBy2-1] ) / 2.0;
    } else {
	$median = $$data_ref[$nBy2];
    }

    return $median;
}


