#!/usr/bin/env perl
#
# original filename: extractStars.pl
#
# Steven Bickerton
#  Dept. of Physics/Astronomy, McMaster University
#  bick@physics.mcmaster.ca
#  Made with makeScript, Mon Jul  9, 2007  14:17:22 DST
#  Host: unknown
#  Working Directory: /Users/bick/working/TASS
#


use strict;
use warnings;
use File::Basename;

use Time::HiRes qw(gettimeofday);

my $exe = basename($0);
my $usage = "Usage: $exe infile n_min\n";

my ($infile, $Nmin) = @ARGV;
die $usage unless $infile and -f $infile;
die $usage unless defined $Nmin;


my $outbase = $infile;
my $outdir = $outbase;
(-d $outdir) || mkdir $outdir;

my $outfile = "$infile";
$outfile .= ".ws";

# column definitions for the .cal file
# needed: $cID $cRA $cDec  $cm1 $cmer1 $cq1  $cm2 $cmer2 $cq2
# unused: $cN  $cjd $cf1 $cf2

my ($cID, $cN, $cRA, $cDec, $cjd, 
    $cf1, $cm1, $cmer1, $cq1, 
    $cf2, $cm2, $cmer2, $cq2) = (0 .. 12);


#my ($cID, $cN, $cRA, $cDec, $cjd, 
#    $cf1, $cm1, $cmer1, $cq1,
#    $cf2, $cm2, $cmer2, $cq2) = ( 0 .. 4, 
#				  12, 5, 6, 10, 
#				  9, 7, 8, 11 );

########################################################################
#
# load the data
#
########################################################################

####################### profiling
my $profile = ($ENV{"TASSPROFILE"}) ? (1) : 0;
my ($tA,$tB,$t0);
$tA = $tB = $t0 = gettimeofday();
printf STDERR "%-20s", "Loading data:" if $profile;
#######################

my %data;
open(INFILE, "$infile");
while (<INFILE>) {
  next if /^#/;
  #  1  2   3    4   5      6     7      8   9     10    11    12   13
  # ID, N, ra, dec, jd, filt1, mag1, merr1, n1, filt2, mag2, merr2, n2
  my @record = (split);
  my $ID = $record[$cID];
  push @{$data{$ID}}, \@record;

}
close(INFILE);



############################################################################
# 
######  go through each ID and get the welch-stetson I statistic  #########
#
############################################################################

###################### profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "I index:" if $profile;
$tA = $tB;
####################

my @output;
IDLOOP: foreach my $ID (keys %data ) {
    
    my $N = scalar @{$data{$ID}};
    next IDLOOP if $N < $Nmin;

    ###############################################################
    # go through and get the weighted mean
    my ($alpha, $delta);
    my ($w1, $w2, $mean1, $mean2, $w1sum, $w2sum, $n1, $n2) = 
	(0,0,0,0,0,0,0,0);
    
  JDLOOP: foreach my $ref (@{$data{$ID}}) {
      # ID, N, ra, dec, jd, filt1, mag1, merr1, q1, filt2, mag2, merr2, q2;
      my ($ra, $dec, $jd, $mag1, $merr1, $q1, $mag2, $merr2, $q2) = 
	  @{$ref}[$cRA,$cDec,$cjd,$cm1,$cmer1,$cq1,$cm2,$cmer2,$cq2];
      
      next JDLOOP
	  if ($merr1 == 0 || $merr2 == 0 || 
	      $mag1 > 25  || $mag2 > 25  || 
	      $q1 || $q2);
      
      my $w1 = 1.0/$merr1**2;
      $mean1 += $w1 * $mag1;
      $w1sum += $w1;
      $n1++;
      my $w2 = 1.0/$merr2**2;
      $mean2 += $w2 * $mag2;
      $w2sum += $w2;
      $n2++;
      
      ($alpha,$delta) = ($ra,$dec);
  }
    
    # n=1 mean=value, n=0 causes div 0 error
    next IDLOOP if ($n1 < 2 || $n2 < 2); 
    
    $mean1 /= $w1sum;
    $mean2 /= $w2sum;
    
    
    ####################################################################
    # now go through again and compute the I index with the new mean
    my ($Isum, $Isum1, $Isum2,   $KsumNumer, $KsumDenom,   $n) = 
	(0.0, 0.0, 0.0,   0.0, 0.0,   0);
  WSLOOP: foreach my $ref (@{$data{$ID}}) {
      # ID, N, ra, dec, jd, filt1, mag1, merr1, q1, filt2, mag2, merr2, q2;
      my ($ra, $dec, $jd, $mag1, $merr1, $q1, $mag2, $merr2, $q2) = 
	  @{$ref}[$cRA,$cDec,$cjd,$cm1,$cmer1,$cq1,$cm2,$cmer2,$cq2];
      
      next WSLOOP
	  if ($merr1 == 0 || $merr2 == 0  ||   $q1 || $q2);
      
      my $Ii = ($mag1 < 25 && $mag2 < 25) ? 
	  ( (($mag1-$mean1)/$merr1) * (($mag2-$mean2)/$merr2) ) : 0.0;
      my $I1i = ($mag1 < 25) ? 
	  ($mag1-$mean1)*($mag1-$mean1)/($merr1*$merr1) : 0.0;
      my $I2i = ($mag2 < 25) ?
	  ($mag2-$mean2)*($mag2-$mean2)/($merr2*$merr2) : 0.0;
      
      #my $delta1 = sqrt($n/($n-1)) * ($mag1-$mean1)/$merr1;
      #my $delta2 = sqrt($n/($n-1)) * ($mag2-$mean2)/$merr2;
      my $delta1 = ($mag1-$mean1)/$merr1;
      my $delta2 = ($mag2-$mean2)/$merr2;
      $KsumNumer += abs( $delta1 ) if ($mag1 < 25);
      $KsumNumer += abs( $delta2 ) if ($mag2 < 25);
      $KsumDenom += $delta1*$delta1 if ($mag1 < 25);
      $KsumDenom += $delta2*$delta2 if ($mag2 < 25);
      
      $Isum += $Ii;
      $Isum1 += $I1i;
      $Isum2 += $I2i;
      
      $n++;
      
  }
    my $I = sqrt(1.0/($n*($n-1))) * $Isum;
    my $I1 = sqrt(1.0/($n*($n-1))) * $Isum1;
    my $I2 = sqrt(1.0/($n*($n-1))) * $Isum2;

    next IDLOOP if $n == 0 or $KsumDenom == 0.0;

    my $K = (1.0/$n) * $KsumNumer / sqrt( (1.0/$n) * $KsumDenom);

    push @output, [$ID, $n1, $alpha, $delta, $mean1, $mean2, $I, $I1, $I2, $K];
    
}




########################################################################
#
#  output
#
#########################################################################

######################## profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
printf STDERR "%-20s", "output:" if $profile;
$tA = $tB;
#######################

open(OUTFILE, ">$outfile");
printf OUTFILE "# 01 ID\n".
    "# 02 N\n".
    "# 03 RA\n".
    "# 04 Dec\n".
    "# 05 Vmean\n".
    "# 06 Imean\n".
    "# 07 Iws\n".
    "# 08 Iws (v only)\n".
    "# 09 Iws (i only)\n".
    "# 10 K (stetson)\n";
foreach my $outref (@output) {
    printf OUTFILE "%d %3d %.4f %.4f  %6.3f %6.3f   %6.3f %6.3f %6.3f  %6.3f\n", @$outref;
}
close(OUTFILE);


######################## profiling
$tB = gettimeofday();
printf STDERR "%7.3f  %7.3f\n", $tB-$tA, $tB-$t0 if $profile;
$tA = $tB;
########################

exit 0;

